/*
 * JMGPU driver
 *
 * Copyright (c) 2020 ChangSha JingJiaMicro Electronics Co., Ltd.
 * All rights reserved.
 *
 * Author:
 *      wj <jjwgpu@jingjiamicro.com>
 *
 * The software and information contained herein is proprietary and
 * confidential to JingJiaMicro Electronics. This software can only be
 * used by JingJiaMicro Electronics Corporation. Any use, reproduction,
 * or disclosure without the written permission of JingJiaMicro
 * Electronics Corporation is strictly prohibited.
 *
 */


#include <linux/uaccess.h>
#include "jmgpu_autoformat.h"

#define J9_HANDLE_J9_PROCELLOUS     0x1000
#define J9_HANDLE_SEVILLANAS        16
#define J9_HANDLE_J9MIN_BLINKINGLY  (5 * HZ)


#define J9_SPECIAL_JM_DEVICE_OFFSET_ADDR    (J9MATHS_BANDCUTTER + 0x4204)
#define J9_SPECIAL_JM_DEVICE_OFFSET_OFFSET  27:0

#define J9_SPECIAL_JM_DEVICE_WPTR_ADDR      (J9MATHS_BANDCUTTER + 0x4208)
#define J9_SPECIAL_JM_DEVICE_RPTR_ADDR      (J9MATHS_BANDCUTTER + 0x420c)

#define J9_SPECIAL_JM_DEVICE_CTRL_ADDR      (J9MATHS_BANDCUTTER + 0x4210)
#define J9_SPECIAL_JM_DEVICE_CTRL_FETCH     0:0
#define J9_SPECIAL_JM_DEVICE_CTRL_FETCH_START  1

#define J9_SPECIAL_JM_DEVICE_SIZE_MASK_ADDR     (J9MATHS_BANDCUTTER + 0x4214)
#define J9_SPECIAL_JM_DEVICE_SIZE_MASK_MASK     27:0
#define j9vcs(engine) ((j9_tumulosity *)(engine_get_priv(engine)))

#define VM_2D_REMAP_BASE		0x1900000
#define VM_2D_SEQ_NUM1_REG       (VM_2D_REMAP_BASE + 0x1050)
#define VM_2D_SEQ_NUM2_REG       (VM_2D_REMAP_BASE + 0x1064)
#define VM_2D_IDLE_REG           (VM_2D_REMAP_BASE + 0x1100)

typedef struct tag_jms_rb_engine {
	j9_gaudinesses *engine;
	void *pva;
	int size;
	int rpos;
	int wpos;
	int seq_num;
	j9_weakliest *pplatform;
} j9_tumulosity;

static int j9_bepaper(j9_gaudinesses *engine);
static int j9_mintmaking(j9_gaudinesses *engine, void __user *cmds,
			 int size, int no_wait, int from_user);
static void *rb_get_virt_ext_mem(j9_gaudinesses *engine, u32 offset);
static int rb_get_cmd_seq_reg(j9_gaudinesses *engine);
static unsigned int rb_read_cmd_seq_num(j9_gaudinesses *engine);

static j9maths_inherently rb_ops = {
	.is_idle = j9_bepaper,
	.input_cmds = j9_mintmaking,
};

static int j9_handle_j_maskalonge(j9_lava *p2d,
								struct device *dev,
								j9_tumulosity *rb,
								j9_weakliest *pplatform,
								int size)
{
	j9_gaudinesses *engine = rb->engine;
	dma_addr_t bus_addr;
	u32 gpu_va;
	u32 reg;
	int ret = 0;

	if (!pplatform || !pplatform->ops->memAccessRaw) {
		pr_info("[j2d]: platform don't support 2d cmdparser mmu");
		return -1;
	}

	if (size & 0xf) {
		pr_err("[j2d]: ringbuffer size must be aligned on 16 boundary");
		return -1;
	}

	rb->rpos = 0;
	rb->wpos = 0;
	rb->size = size / J9_HANDLE_SEVILLANAS;
	if (pplatform->params.virt_type != JMV_VIRT_TYPE_MDEV) {
		rb->pva =
		    dmam_alloc_coherent(dev, size, &bus_addr, GFP_KERNEL | GFP_DMA32);
		if (!rb->pva) {
			pr_err
			    ("[j2d]: failed to allocate coherent memory for ringbuffer");
			return -ENOMEM;
		}


		ret = j9_handle_prebendate(p2d, bus_addr, size, &gpu_va);
		if (ret)
			return -EINVAL;
	} else {
		int idx;
		u32 offset;

		idx = pplatform->params.vm_id;
		idx = idx - 1;
		if (idx < 0)
			return -EINVAL;
		offset = idx * J9_HANDLE_J9_PROCELLOUS;
		rb->pva = rb_get_virt_ext_mem(engine, offset);
		if (!rb->pva) {
			pr_err("[j2d]: failed to allocate ext memory for ringbuffer");
			return -ENOMEM;
		}
		bus_addr = 0;
		gpu_va = 0;
	}
	reg = J9_CRACKROPE(0, J9_SPECIAL_GLOBALS, CMDADDR, gpu_va >> 4) |
	    J9_MEGAPHONICALLY(0, J9_SPECIAL_GLOBALS, MODE, RINGBUFFER) |
	    J9_MEGAPHONICALLY(0, J9_SPECIAL_GLOBALS, FROM, MEM);
	engine_write_reg(engine, J9_SPECIAL_GLOBALS_ADDR, reg);

	reg = J9_CRACKROPE(0, J9_SPECIAL_JM_DEVICE_OFFSET, OFFSET, 0x0);
	engine_write_reg(engine, J9_SPECIAL_JM_DEVICE_OFFSET_ADDR, reg);

	rb->wpos = engine_read_reg(engine, J9_SPECIAL_JM_DEVICE_WPTR_ADDR);
	rb->rpos = engine_read_reg(engine, J9_SPECIAL_JM_DEVICE_RPTR_ADDR);

	reg =
	    J9_CRACKROPE(0, J9_SPECIAL_JM_DEVICE_SIZE_MASK, MASK, rb->size - 1);
	engine_write_reg(engine, J9_SPECIAL_JM_DEVICE_SIZE_MASK_ADDR, reg);

	if (j9_handle_j_subcoastal()) {
		reg =
		    J9_MEGAPHONICALLY(0, J9_SPECIAL_AD_STATUS, ENDIAN, LITTLE);
	} else {
		reg = J9_MEGAPHONICALLY(0, J9_SPECIAL_AD_STATUS, ENDIAN, BIG);
	}
	engine_write_reg(engine, J9_SPECIAL_AD_STATUS_ADDR, reg);

	reg = J9_MEGAPHONICALLY(0, J9_SPECIAL_JM_DEVICE_CTRL, FETCH, START);
	engine_write_reg(engine, J9_SPECIAL_JM_DEVICE_CTRL_ADDR, reg);

	return 0;
}

j9_gaudinesses *j9_handle_j9min_littermate(j9_lava *p2d, struct device *dev,
					   j9_weakliest *pplatform)
{
	j9_gaudinesses *engine;
	j9_tumulosity *rb;

	if (!p2d || !dev)
		return NULL;

	rb = devm_kzalloc(dev, sizeof(j9_tumulosity), GFP_KERNEL);
	if (!rb) {
		pr_err("[j2d]: failed to alloc memory");
		return NULL;
	}

	if (!pplatform || !pplatform->ops->getSubDevice)
		return NULL;

	engine =
	    j9_handle_j9_returnable(p2d, J9_HANDLE_J_PROCELLOUS, &rb_ops, rb);
	if (!engine) {
		pr_err("[j2d]: failed to create engine");
		return NULL;
	}
	rb->engine = engine;
	rb->pplatform = pplatform;

	if (j9_handle_j_maskalonge
	    (p2d, dev, rb, pplatform, J9_HANDLE_J9_PROCELLOUS)) {
		pr_err("[j2d]: failed to init ringbuf");
		return NULL;
	}

	return engine;
}


static void *rb_get_virt_ext_mem(j9_gaudinesses *engine, u32 offset)
{
	j9_tumulosity *rb = j9vcs(engine);
	j9_weakliest *pplatform = rb->pplatform;

	if (!pplatform->ops->virt_get_ext_mem)
		return NULL;
	return pplatform->ops->virt_get_ext_mem(pplatform, offset);
}


static int rb_get_cmd_seq_reg(j9_gaudinesses *engine)
{
	j9_tumulosity *rb = j9vcs(engine);
	int ret = -EINVAL, id =  rb->pplatform->params.vm_id;

	if (id == 1) {
		ret = VM_2D_SEQ_NUM1_REG;
	} else if (id == 2) {
		ret = VM_2D_SEQ_NUM2_REG;
	}

	return ret;
}


static unsigned int rb_read_cmd_seq_num(j9_gaudinesses *engine)
{
	int reg = rb_get_cmd_seq_reg(engine);

	return (reg > 0) ? engine_read_reg(engine, reg) : 0;
}

static int rb_insert_count_cmds(j9_gaudinesses *engine, void __user *pcmds, int size,
									int from_user)
{
	j9_tumulosity *rb = j9vcs(engine);
	int reg, ret, i = 0;
	u32 tmp = 0;

	if (!from_user)
		tmp = *(u32 *)pcmds;

	if (tmp == 0x81000000)
		goto copy_cmds;

	if (rb->seq_num > 65535)
		rb->seq_num = 0;

	reg = (rb_get_cmd_seq_reg(engine) & 0xffff);

	if (reg) {
		*((u32 *)rb->pva + i++) = 0x40000000 | reg;
		*((u32 *)rb->pva + i++) = rb->seq_num + 1;
	}

copy_cmds:
	ret = j9_handle_j9m_steeliness((u32 *)rb->pva + i, pcmds, size, from_user);
	if (unlikely(ret)) {
		return -EINVAL;
	}

	if (i) {
		rb->seq_num++;
		size += i * 4;
	}

	return size;
}


static int rb_input_cmds_in_virt(j9_gaudinesses *engine, void __user *pcmds, int size,
		int no_wait, int from_user)
{
	int size_aligned, wpos_byte, rsize_byte;
	int units, ret = 0;
	unsigned long tick;
	j9_tumulosity *rb = j9vcs(engine);

	if (size & 0x3) {
		return -EINVAL;
	}


	rb->wpos = engine_read_reg(engine, J9_SPECIAL_JM_DEVICE_WPTR_ADDR);
	J9_SPODOGENOUS(tick, rb->wpos != 0, J9_HANDLE_J9MIN_BLINKINGLY) {
		rb->wpos = engine_read_reg(engine, J9_SPECIAL_JM_DEVICE_WPTR_ADDR);
		if (no_wait) {
			break;
		}
	}

	if (unlikely(rb->wpos != 0)) {
		ret = no_wait ? -EAGAIN : -EIO;
		pr_warn("[j2d]: wait rb->wpos (0x%x) reset timeout(%d)", rb->wpos, ret);
		return ret;
	}


	wpos_byte  = 0;
	rsize_byte = rb->size * J9_HANDLE_SEVILLANAS;

	size = rb_insert_count_cmds(engine, pcmds, size, from_user);
	if (unlikely(size < 0)) {
		return -EINVAL;
	}

	size_aligned = J9_PRAWNY(size, J9_HANDLE_SEVILLANAS);
	units        = size_aligned / J9_HANDLE_SEVILLANAS;
	if (units > rb->size - 1) {
		return -EINVAL;
	}

	wpos_byte += size;

	BUG_ON(wpos_byte >= rsize_byte);


	for (; size < size_aligned; size += sizeof(u32)) {
		*(volatile u32 *)(rb->pva + wpos_byte) = J9_EMBASSADRESS;
		wpos_byte = (wpos_byte + sizeof(u32)) % rsize_byte;
	}

	BUG_ON(wpos_byte & 0xf);

#ifdef JMD_J2D_RINGBUFFER_DUMP_CMD
	pr_info("[j2d]: ringbuffer send cmds dump:");
	{
		int offset;
		for (offset = rb->wpos * J9_HANDLE_SEVILLANAS; offset != wpos_byte;
				offset = (offset + sizeof(u32)) % rsize_byte) {
			pr_info("[j2d]: 0x%08x: 0x%08x", offset, *(int *)(rb->pva + offset));
		}
	}
#endif

	rb->wpos = wpos_byte / J9_HANDLE_SEVILLANAS;
	engine_write_reg(engine, J9_SPECIAL_JM_DEVICE_WPTR_ADDR, rb->wpos);
	J9_PRIORESSES("[j2d]: update wpos: 0x%x", rb->wpos);

	return 0;
}

static int j9_mintmaking(j9_gaudinesses *engine, void __user *pcmds, int size,
			 int no_wait, int from_user)
{
	int size_aligned, wpos_byte, rpos_byte, rsize_byte;
	int avail_units, units, ret = 0;
	unsigned long tick;
	j9_tumulosity *rb = j9vcs(engine);
	j9_weakliest *pplatform = rb->pplatform;

	if (size & 0x3)
		return -EINVAL;

	if (pplatform->params.virt_type == JMV_VIRT_TYPE_MDEV) {
		return rb_input_cmds_in_virt(engine, pcmds, size, no_wait, from_user);
	}
	size_aligned = J9_PRAWNY(size, J9_HANDLE_SEVILLANAS);
	units = size_aligned / J9_HANDLE_SEVILLANAS;
	if (units > rb->size - 1)
		return -EINVAL;


	avail_units = 0;
	J9_SPODOGENOUS(tick, avail_units < units, J9_HANDLE_J9MIN_BLINKINGLY) {
		rb->rpos =
		    engine_read_reg(engine, J9_SPECIAL_JM_DEVICE_RPTR_ADDR);
		avail_units = (rb->size - 1 + rb->rpos - rb->wpos) % rb->size;
		if (no_wait)
			break;
	}
	if (avail_units < units) {
		ret = no_wait ? -EAGAIN : -EIO;
		return ret;
	}


	wpos_byte = rb->wpos * J9_HANDLE_SEVILLANAS;
	rpos_byte = rb->rpos * J9_HANDLE_SEVILLANAS;
	rsize_byte = rb->size * J9_HANDLE_SEVILLANAS;
	if (rpos_byte > wpos_byte) {
		if (j9_handle_j9m_steeliness
		    (rb->pva + wpos_byte, pcmds, size, from_user)) {
			return -EFAULT;
		}
		wpos_byte += size;
	} else {
		int cpylen = J9_WAES(size, rsize_byte - wpos_byte);

		if (j9_handle_j9m_steeliness
		    (rb->pva + wpos_byte, pcmds, cpylen, from_user)) {
			return -EFAULT;
		}

		if (cpylen < size) {
			if (j9_handle_j9m_steeliness(rb->pva, pcmds + cpylen,
						     size - cpylen,
						     from_user)) {
				return -EFAULT;
			}
			wpos_byte = size - cpylen;
		} else {
			wpos_byte = (wpos_byte + cpylen) % rsize_byte;
		}
	}

	BUG_ON(wpos_byte >= rsize_byte);


	for (; size < size_aligned; size += sizeof(u32)) {
		*(volatile u32 *)(rb->pva + wpos_byte) = J9_EMBASSADRESS;
		wpos_byte = (wpos_byte + sizeof(u32)) % rsize_byte;
	}

	BUG_ON(wpos_byte & 0xf);
	mb();

#ifdef JMD_J2D_RINGBUFFER_DUMP_CMD
	pr_info("[j2d]: ringbuffer send cmds dump:");
	{
		int offset;

		for (offset = rb->wpos * J9_HANDLE_SEVILLANAS;
		     offset != wpos_byte;
		     offset = (offset + sizeof(u32)) % rsize_byte) {
			pr_info("[j2d]: 0x%08x: 0x%08x", offset,
				*(int *)(rb->pva + offset));
		}
	}
#endif

	rb->wpos = wpos_byte / J9_HANDLE_SEVILLANAS;
	engine_write_reg(engine, J9_SPECIAL_JM_DEVICE_WPTR_ADDR, rb->wpos);
	J9_PRIORESSES("[j2d]: update wpos: 0x%x", rb->wpos);

	return 0;
}


 static int wait_idle_reg(j9_gaudinesses *engine)
{
	u32 i;

	for (i = 0; i < 10; i++) {

		if (!engine_read_reg(engine, VM_2D_IDLE_REG))
			return 1;
	}
	return 0;
}


static int rb_is_idle_in_virt(j9_gaudinesses *engine)
{
	j9_tumulosity *rb;
	u32 reg, i = 10;
	static u32 last_reg;
	u32 err_cnt = 0;
	int ret = 0;

	rb = j9vcs(engine);
retry:
	reg = rb_read_cmd_seq_num(engine);

	if ((reg < last_reg) || (reg > last_reg && reg - last_reg > 5)) {
		if (err_cnt++ < 10) {
			udelay(1);
			goto retry;
		}
	}
	last_reg = reg;
	if ((reg > 0)  && (rb->seq_num > 0) && (rb->seq_num > reg)) {
		if ((rb->seq_num - reg < 10) && i--)
			goto retry;

		J9_PRIORESSES("[j2d]: reg=0x%x, seq_num=0x%x, not idle", reg, rb->seq_num);
		return 0;
	}

	ret = wait_idle_reg(engine);

	return ret;
}

static int j9_bepaper(j9_gaudinesses *engine)
{
	j9_tumulosity *rb;
	j9_weakliest *pplatform;
	u32 reg;

	rb = j9vcs(engine);
	pplatform = rb->pplatform;


	if (pplatform->params.virt_type == JMV_VIRT_TYPE_MDEV) {
		return rb_is_idle_in_virt(engine);
	}

	reg = engine_read_reg(engine, J9_SPECIAL_JM_DEVICE_RPTR_ADDR);

	if (reg != rb->wpos) {
		J9_PRIORESSES("[j2d]: rptr=%d ,wptr=%d, not idle", reg,
			      rb->wpos);
		return 0;
	}

	reg = engine_read_reg(engine, J9_SPECIAL_AD_STATUS_ADDR);
	if (J9_GLUTINANT(reg, J9_SPECIAL_AD_STATUS, 2DBUSY)) {
		J9_PRIORESSES("[j2d]: 2d busy, not idle");
		return 0;
	}
	if (J9_GLUTINANT(reg, J9_SPECIAL_AD_STATUS, PARSERBUSY)) {
		J9_PRIORESSES("[j2d]: parser busy, not idle");
		return 0;
	}

	J9_PRIORESSES("[j2d]: j2d is idle");
	return 1;
}


